#!/usr/bin/env python2
import unittest
import rospy
from rospy.exceptions import ROSException
from std_msgs.msg import Bool
import std_msgs.msg
from std_srvs.srv import Empty, EmptyRequest
from gazebo_msgs.msg import ModelStates
from geometry_msgs.msg import TwistStamped
from dataspeed_ulc_msgs.msg import UlcCmd, UlcReport


class GazeboHandshake:
    def __init__(self):
        self.received_model_states = False

    def recv_model_states(self, msg):
        self.received_model_states = True

    def start(self):
        rospy.Subscriber('/gazebo/model_states', ModelStates, self.recv_model_states)

        srv = rospy.ServiceProxy('/start_gazebo', Empty)
        try:
            srv.wait_for_service(timeout=10)
        except ROSException as e:
            return False

        srv.call(EmptyRequest())

        rospy.loginfo('waiting for Gazebo to become active')
        num_tries = 3
        for i in range(0, num_tries):
            timeout_t = rospy.Time.now() + rospy.Duration(5)
            while not rospy.is_shutdown() and not self.received_model_states and (timeout_t - rospy.Time.now()).to_sec() > 0:
                rospy.rostime.wallsleep(0.05)

            if not self.received_model_states:
                rospy.loginfo('Gazebo not active yet (attempt #%d)... Triggering restart' % (i + 1))
                rospy.ServiceProxy('/stop_gazebo', Empty).call(EmptyRequest())
                rospy.rostime.wallsleep(0.5)
                rospy.ServiceProxy('/start_gazebo', Empty).call(EmptyRequest())
            else:
                rospy.loginfo('Gazebo found active after %f seconds' % (rospy.Time.now() - timeout_t).to_sec())
                return True

        return False

    @staticmethod
    def stop():
        rospy.ServiceProxy('/stop_gazebo', Empty).call(EmptyRequest())


class VehicleTest(unittest.TestCase, GazeboHandshake):
    def setUp(self):
        GazeboHandshake.__init__(self)
        self.ulc_report = None
        self.twist = None
        self.enabled = False
        gazebo_active = self.start()
        self.assertTrue(gazebo_active, msg='Gazebo failed to start')

        self.pub_cmd = rospy.Publisher('ulc_cmd', UlcCmd, queue_size=1)
        self.pub_enable = rospy.Publisher('enable', std_msgs.msg.Empty, queue_size=1, latch=True)
        rospy.Subscriber('ulc_report', UlcReport, self.recv_ulc_report)
        rospy.Subscriber('twist', TwistStamped, self.recv_twist)
        rospy.Subscriber('dbw_enabled', Bool, self.recv_enable)

    def tearDown(self):
        self.stop()
        rospy.rostime._set_rostime(rospy.Time(0))
        pass

    def recv_ulc_report(self, msg):
        self.ulc_report = msg

    def recv_twist(self, msg):
        self.twist = msg

    def recv_enable(self, msg):
        self.enabled = msg.data

    def publish_ulc_cmd(self):
        self.pub_cmd.publish(UlcCmd(
            linear_velocity=self.target_speed,
            yaw_command=self.target_yaw_rate,
            enable_shifting=True,
            enable_steering=True,
            shift_from_park=True,
            enable_pedals=True
        ))

    def test_vehicle_plugin(self):
        time_resolution = 0.02

        wait_for_time = rospy.get_time()
        while wait_for_time == 0.0:
            wait_for_time = rospy.get_time()

        self.target_speed = 50.0 * 0.447  # mph to m/s
        self.target_yaw_rate = 0.2
        self.speed_tolerance = 1.0 * 0.447
        try:
            rospy.loginfo('Checking for received ULC and twist report messages')
            timeout_t = rospy.Time.now() + rospy.Duration(15.0)
            while not rospy.is_shutdown() and ((timeout_t - rospy.Time.now()).to_sec() > 0) and (self.ulc_report is None or self.twist is None):
                self.publish_ulc_cmd()
                rospy.sleep(time_resolution)

            self.assertTrue(self.ulc_report is not None,
                            msg='Did not receive ULC report message in 15 seconds')

            self.assertTrue(self.twist is not None,
                            msg='Did not receive twist message in 15 seconds')

            rospy.loginfo('Enable DBW')
            self.pub_enable.publish(std_msgs.msg.Empty())
            rospy.sleep(3.0)
            self.assertTrue(self.enabled, msg='Drive-by-wire not enabled')

            rospy.loginfo('Waiting for reference to reach target')
            timeout_t = rospy.Time.now() + rospy.Duration(90)
            while not rospy.is_shutdown() and (timeout_t - rospy.Time.now()).to_sec() > 0 and abs(self.ulc_report.speed_ref - self.target_speed) > 0.03:
                self.publish_ulc_cmd()
                rospy.sleep(time_resolution)

            self.assertAlmostEqual(
                first=self.ulc_report.speed_ref - self.target_speed,
                second=0.0,
                delta=0.03,
                msg='Speed reference never made it to target speed (target = %f, ref = %f)' % (self.target_speed, self.ulc_report.speed_ref)
            )

        except rospy.ROSInterruptException:
            self.fail(msg='ROS shutdown before test complete')


if __name__ == '__main__':
    import rostest
    rospy.init_node('vehicle_plugin_test')
    rostest.rosrun('dataspeed_dbw_gazebo', 'vehicle_plugin_test', VehicleTest)