#!/usr/bin/env python

import rospy
import time
import threading
from select import select
import evdev
from evdev import InputDevice, list_devices, ecodes, ff, _uinput
from sensor_msgs.msg import Joy, JoyFeedback, JoyFeedbackArray

g_joy_msg = Joy()


class DeviceInterface:
    def __init__(self):
        self.buttons = {}
        self.axes = {}
        self.axis_info = {}
        self.device = None
        self.connected = False
        self.constant_force_effect = None

    def __init_constant_effect(self, duration):
        self.__upload_constant_effect(-1, duration, 0)

    def __update_constant_effect(self, duration, level):
        self.__upload_constant_effect(self.constant_force_effect.id, duration, level)

    def __upload_constant_effect(self, id, duration, level):
        self.constant_force_effect = ff.Effect(
            type=ecodes.FF_CONSTANT,
            id=id,
            direction=0x4000, # Left
            ff_trigger=ff.Trigger(0, 0),
            ff_replay=ff.Replay(duration, 0),
            u=ff.EffectType(
                ff_constant_effect=ff.Constant(
                    level=level,
                    ff_envelope=ff.Envelope(
                        attack_length=1,
                        attack_level=level,
                        fade_length=1,
                        fade_level=level
                    )
                )
            )
        )
        self.constant_force_effect.id = self.device.upload_effect(self.constant_force_effect)

    def __play_effect(self, id):
        _uinput.write(self.device.fileno(), ecodes.EV_FF, id, 1)

    def open_device(self, name):
        dev_paths = list_devices()
        if len(name) == 0:
            pass
        else:
            for d in map(InputDevice, list_devices()):
                if name in d.name:
                    self.device = InputDevice(d.fn)
                    rospy.loginfo('ff_joy: Connected to device [%s]' % d.name)

                    # Get device capability information
                    button_capability = self.device.capabilities().get(ecodes.EV_KEY)
                    axis_capability = self.device.capabilities().get(ecodes.EV_ABS)
                    feedback_capability = self.device.capabilities().get(ecodes.EV_FF)

                    # Resize button array to match number of supported key events
                    g_joy_msg.buttons = len(button_capability) * [0]
                    for idx in button_capability:
                        self.buttons[idx] = 0

                    # Resize axes array to match number of supported absolute events
                    g_joy_msg.axes = len(axis_capability) * [float('nan')]
                    for idx, info in axis_capability:
                        self.axes[idx] = float('nan')
                        self.axis_info[idx] = {'min': float(info.min), 'max': float(info.max), 'range': float(info.max - info.min)}

                    # Upload constant force feedback event if supported by device
                    if ecodes.FF_CONSTANT in feedback_capability:
                        rospy.loginfo('ff_joy: Detected constant force feedback capability for [%s]' % name)
                        self.__init_constant_effect(duration=100)

                    self.connected = True
                    break
            if not self.connected:
                if len(dev_paths) == 0:
                    rospy.logwarn_throttle(10, 'ff_joy: No available devices')
                else:
                    rospy.logwarn('ff_joy: Failed to find device [%s]\nAvailable devices:' % name)
                    for path in dev_paths:
                        rospy.logwarn('  %s' % InputDevice(path).name)

    def process_events(self):
        select([self.device], [], [], 0.05)
        try:
            for event in self.device.read():
                if event.type == ecodes.EV_KEY:
                    self.buttons[event.code] = event.value
                elif event.type == ecodes.EV_ABS:
                    min_val = self.axis_info[event.code]['min']
                    if min_val == 0:
                        self.axes[event.code] = event.value / self.axis_info[event.code]['max']
                    else:
                        self.axes[event.code] = 2 * (event.value - min_val) / self.axis_info[event.code]['range'] - 1.0

                # Populate buttons and axes in ROS message
                for i in range(0, len(g_joy_msg.buttons)):
                    g_joy_msg.buttons[i] = self.buttons.values()[i]
                for i in range(0, len(g_joy_msg.axes)):
                    g_joy_msg.axes[i] = self.axes.values()[i]
        except IOError as e:
            if e.errno == 19:
                self.connected = False
                rospy.logwarn_throttle(2, 'ff_joy: Lost connection to [%s]... Will try to reconnect every second' % self.device.name)

    def send_feedback(self, feedback):
        if self.connected:
            if self.constant_force_effect is not None:
                self.__update_constant_effect(
                    duration=100,
                    level=int(feedback.intensity * 0x7FFF)
                )
                self.__play_effect(self.constant_force_effect.id)
            else:
                rospy.logwarn_throttle(2, 'ff_joy: Trying to send constant force feedback, but it is not supported by the device')


class JoystickEventThread (threading.Thread):
    def __init__(self, dev_name):
        threading.Thread.__init__(self)
        self.stop_event = threading.Event()
        self.device_interface = DeviceInterface()
        self.dev_name = dev_name

    def run(self):
        if self.dev_name == '':
            dev_list = [InputDevice(path) for path in list_devices()]
            rospy.logwarn('[dev] parameter not set; exiting')
            if len(dev_list) == 0:
                rospy.logwarn('No devices available')
            else:
                rospy.loginfo('Available devices:')
                for d in dev_list:
                    rospy.loginfo('  %s' % d.name)
            rospy.signal_shutdown('')
            return

        while not self.stop_event.isSet():
            if not self.device_interface.connected:
                self.device_interface.open_device(self.dev_name)
                if self.device_interface.connected:
                    continue
                time.sleep(1)
            else:
                self.device_interface.process_events()

    def stop(self):
        self.stop_event.set()


class FFJoyNode:
    def __init__(self):
        rospy.init_node('ff_joy')

        pub_rate = rospy.get_param('~pub_rate', default=100.0)
        dev_name = rospy.get_param('~dev', default='')
        self.pub_joy = rospy.Publisher('joy', Joy, queue_size=1)
        rospy.Subscriber('joy_feedback', JoyFeedbackArray, self.recv_feedback)

        rospy.Timer(rospy.Duration(1.0 / pub_rate), self.joy_timer_cb)

        self.joystick_thread = JoystickEventThread(dev_name)
        self.joystick_thread.start()

    def joy_timer_cb(self, event):
        if self.joystick_thread.device_interface.connected:
            g_joy_msg.header.stamp = event.current_real
            try:
                self.pub_joy.publish(g_joy_msg)
            except rospy.ROSException as e:
                rospy.logwarn_throttle(0.5, '[ff_joy]: %s' % str(e))

    def recv_feedback(self, msg):
        for feedback in msg.array:
            if feedback.type != 3:
                rospy.logwarn_throttle(0.5, 'Unsupported feedback type [%d], driver only supports [3]')
            else:
                self.joystick_thread.device_interface.send_feedback(feedback)

    def shutdown_handler(self):
        try:
            self.joystick_thread.stop()
            self.joystick_thread.join()
        except RuntimeError:
            pass


if __name__ == "__main__":
    node_instance = FFJoyNode()
    rospy.on_shutdown(node_instance.shutdown_handler)
    rospy.spin()
